"""Vault specific code"""

import base64
import json
from datetime import datetime, timezone
from typing import Any, Final

import requests
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import BaseModel, ValidationError

from .helpers import b64str, prepare_gh_app_jwt, private_pem_to_der, vault_wrap_key


class VaultAPIError(Exception):
    """Any error response from the Vault API"""


class AppKeyImportError(VaultAPIError):
    """Failure to upload/import the wrapped GitHub App key into Vault"""


class JWTSigningError(VaultAPIError):
    """Failure to have Vault sign a GitHub App JWT token"""


class TokenRevokeError(VaultAPIError):
    """Failure to self-revoke the Vault token"""


class WrappingKeyDownloadError(VaultAPIError):
    """Failure to download the Vault Transit wrapping key"""


class VaultErrors(BaseModel):
    """
    https://developer.hashicorp.com/vault/api-docs#error-response
    """

    errors: list[str]


class JWTData(BaseModel):
    """Part of SignedJWT"""

    signature: str


class SignedJWT(BaseModel):
    """
    https://developer.hashicorp.com/vault/api-docs/secret/transit#sign-data
    """

    data: JWTData


class KeyData(BaseModel):
    """Part of WrappingKey"""

    public_key: str


class WrappingKey(BaseModel):
    """
    https://developer.hashicorp.com/vault/api-docs/secret/transit#get-wrapping-key
    """

    data: KeyData


class VaultTransit:
    """Interact with Vault's Transit Secrets Engine"""

    def __init__(self, vault_addr: str, vault_token: str, transit_backend: str):
        """
        :param vault_addr: Vault instance VAULT_ADDR.
        :param vault_token: Vault instance VAULT_TOKEN.
        :param transit_backend: Transit backend mount path.
        """
        self.vault_addr: Final[str] = vault_addr.rstrip("/")
        self.auth_headers: Final[dict[str, str]] = {"X-Vault-Token": vault_token}
        self.transit_backend: Final[str] = transit_backend.strip("/")

    def __download_wrapping_key(self) -> rsa.RSAPublicKey:
        wrapping_key_url = self.vault_addr + f"/v1/{self.transit_backend}/wrapping_key"

        try:
            response = requests.get(
                wrapping_key_url,
                headers=self.auth_headers,
                timeout=10,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            try:
                errors_bm = VaultErrors(**http_error.response.json())
                error_message = "\n".join(errors_bm.errors)
            except Exception:  # pylint: disable=broad-exception-caught
                error_message = "<Failed to parse Vault API error response>"
            raise WrappingKeyDownloadError(error_message) from http_error

        try:
            wrapping_key_bm = WrappingKey(**response.json())
        except ValidationError as validation_error:
            error_message = "<Failed to parse Wrapping Key API response>"
            raise WrappingKeyDownloadError(error_message) from validation_error

        wrapping_pem_key = wrapping_key_bm.data.public_key.encode()
        wrapping_key = serialization.load_pem_public_key(wrapping_pem_key)

        if not isinstance(wrapping_key, rsa.RSAPublicKey):
            raise ValueError("Unexpected wrappingkey format")

        return wrapping_key

    def __api_write(
        self,
        api_path: str,
        payload: None | dict[str, Any] = None,
        vault_exception: type[Exception] = VaultAPIError,
    ) -> requests.models.Response:
        update_url = self.vault_addr + api_path

        if payload is None:
            payload = {}

        try:
            response = requests.post(
                update_url,
                headers=self.auth_headers,
                data=json.dumps(payload),
                timeout=10,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            try:
                errors_bm = VaultErrors(**http_error.response.json())
                error_message = "\n".join(errors_bm.errors)
            except Exception:  # pylint: disable=broad-exception-caught
                error_message = "<Failed to parse Vault API error response>"
            raise vault_exception(error_message) from http_error

        return response

    def import_key(self, key_name: str, pem_app_key: bytes) -> None:
        """
        Import GitHub App key

        :param key_name: Name the Transit Engine will know the key by.
        :param pem_app_key: The App's PEM formated private RSA key.
        """
        der_app_key: bytes = private_pem_to_der(pem_app_key)
        wrapping_key: rsa.RSAPublicKey = self.__download_wrapping_key()
        wrapped_b64: str = vault_wrap_key(der_app_key, wrapping_key)

        api_path = f"/v1/{self.transit_backend}/keys/{key_name}/import"
        payload = {
            "ciphertext": wrapped_b64,
            "hash_function": "SHA256",
            "type": "rsa-2048",
            "exportable": False,
            "allow_plaintext_backup": False,
        }

        self.__api_write(api_path, payload, AppKeyImportError)

    def sign_jwt(self, key_name: str, app_id: str) -> str:
        """
        Sign JWT token to authenticate towards GitHub

        :param key_name: Transit Engine key name.
        :param app_id: GitHub App ID.


        :return: GitHub App JWT token
        """

        now = datetime.now(timezone.utc)
        header_and_claims = prepare_gh_app_jwt(app_id, now)

        api_path = f"/v1/{self.transit_backend}/sign/{key_name}"
        payload = {
            "input": b64str(header_and_claims),
            "hash_algorithm": "sha2-256",
            "signature_algorithm": "pkcs1v15",
        }

        response: requests.models.Response = self.__api_write(
            api_path, payload, JWTSigningError
        )

        try:
            signature_bm = SignedJWT(**response.json())
        except ValidationError as validation_error:
            error_message = "<Failed to parse Sign JWT API response>"
            raise JWTSigningError(error_message) from validation_error

        signature = signature_bm.data.signature.removeprefix("vault:v1:")
        signature = b64str(base64.b64decode(signature), urlsafe=True)

        jwt_token = header_and_claims + "." + signature
        return jwt_token

    def revoke_token(self) -> None:
        """Vault Token self-revoke"""

        api_path = "/v1/auth/token/revoke-self"
        self.__api_write(api_path, vault_exception=TokenRevokeError)



"""GitHub specific code"""

import json
from datetime import datetime
from typing import Final, Literal, TypedDict

import requests
from pydantic import BaseModel, Field, TypeAdapter, ValidationError

PermARW = None | Literal["admin", "read", "write"]
PermRW = None | Literal["read", "write"]
PermR = None | Literal["read"]
PermW = None | Literal["write"]


class TokenResponse(TypedDict, total=False):
    """Typing for customized Access Token response"""

    access_token: str
    expires_at: datetime
    permissions: dict[str, str]
    repositories: list[str]


class GitHubAPIError(Exception):
    """Any error response from the GitHub API"""


class InstallationLookupError(GitHubAPIError):
    """Failure to lookup the GitHub App installation ID"""


class TokenIssueError(GitHubAPIError):
    """Failure to issue GitHub Access Token"""


class NotInstalledError(Exception):
    """The GitHub App isn't installed in the specified account"""


class GitHubErrors(BaseModel):
    """
    https://docs.github.com/en/rest/overview/resources-in-the-rest-api?apiVersion=2022-11-28
    """

    message: str


class AccountInfo(BaseModel):
    """Part of Installation"""

    login: str = Field(
        max_length=39, pattern=r"^[a-zA-Z0-9]([a-zA-Z0-9\-]*[a-zA-Z0-9])?$"
    )


class Installation(BaseModel):
    """
    https://docs.github.com/en/rest/apps/apps?apiVersion=2022-11-28#list-installations-for-the-authenticated-app
    """

    id: int
    account: AccountInfo


class TokenPermissions(BaseModel):
    """Part of AccessToken"""

    # Repository permissions
    actions: PermRW = None
    administration: PermRW = None
    checks: PermRW = None
    contents: PermRW = None
    deployments: PermRW = None
    environments: PermRW = None
    issues: PermRW = None
    metadata: PermRW = None
    packages: PermRW = None
    pages: PermRW = None
    pull_requests: PermRW = None
    repository_hooks: PermRW = None
    repository_projects: PermARW = None
    secret_scanning_alerts: PermRW = None
    secrets: PermRW = None
    security_events: PermRW = None
    single_file: PermRW = None
    statuses: PermRW = None
    vulnerability_alerts: PermRW = None
    workflows: PermW = None
    # Organizational permissions
    members: PermRW = None
    organization_administration: PermRW = None
    organization_custom_roles: PermRW = None
    organization_announcement_banners: PermRW = None
    organization_hooks: PermRW = None
    organization_personal_access_tokens: PermRW = None
    organization_personal_access_token_requests: PermRW = None
    organization_plan: PermR = None
    organization_projects: PermARW = None
    organization_packages: PermRW = None
    organization_secrets: PermRW = None
    organization_self_hosted_runners: PermRW = None
    organization_user_blocking: PermRW = None
    team_discussions: PermRW = None


class Repository(BaseModel):
    """Part of AccessToken"""

    name: str = Field(max_length=100, pattern=r"^[a-zA-Z0-9_\-\.]+$")


class AccessToken(BaseModel):
    """
    https://docs.github.com/en/rest/apps/apps?apiVersion=2022-11-28#create-an-installation-access-token-for-an-app
    """

    token: str
    expires_at: datetime
    permissions: TokenPermissions
    repositories: None | list[Repository] = None


class GitHubApp:
    """GitHub App Access Tokens, etc"""

    def __init__(self, account: str, jwt_token: str):
        """
        :param app_id: GitHub App ID.
        :param jwt_token: GitHub App JWT token
        """

        self.account: Final[str] = account
        self.auth_headers: Final[dict[str, str]] = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {jwt_token}",
            "X-GitHub-Api-Version": "2022-11-28",
        }

    def __find_installation(self) -> str:
        lookup_url = "https://api.github.com/app/installations"

        pagination_params = {
            "page": 1,
            "per_page": 100,
        }

        more = True
        while more:
            more = False
            try:
                response = requests.get(
                    lookup_url,
                    headers=self.auth_headers,
                    params=pagination_params,
                    timeout=10,
                )
                response.raise_for_status()
            except requests.exceptions.HTTPError as http_error:
                try:
                    errors_bm = GitHubErrors(**http_error.response.json())
                    error_message = errors_bm.message
                except Exception:  # pylint: disable=broad-exception-caught
                    error_message = "<Failed to parse GitHub API error response>"
                raise InstallationLookupError(error_message) from http_error

            try:
                ita = TypeAdapter(list[Installation])
                installations = ita.validate_python(response.json())
            except ValidationError as validation_error:
                error_message = "<Failed to parse Installations API response>"
                raise InstallationLookupError(error_message) from validation_error

            for installation in installations:
                if installation.account.login.lower() == self.account.lower():
                    return str(installation.id)

            if "next" in response.links.keys():
                pagination_params["page"] += 1
                more = True

        failure = f'App appear not to be installed in the "{self.account}" account'
        raise NotInstalledError(failure)

    def issue_token(
        self,
        permissions: None | dict[str, str] = None,
        repositories: None | list[str] = None,
    ) -> TokenResponse:
        """
        Issue GitHub Access Token

        :param permissions: Optionally scope (down) token permissions.
        :param repositories: Optionally limit accessible repositories.

        :return: The requested access token; together with its expiry
            time, permission scope and optionally covered repositories.
        """

        params: dict[str, dict[str, str] | list[str]] = {}
        if permissions:
            params.update({"permissions": permissions})
        if repositories:
            params.update({"repositories": repositories})

        installation_id: str = self.__find_installation()
        issue_url = "/".join(
            [
                "https://api.github.com/app/installations",
                installation_id,
                "access_tokens",
            ]
        )

        try:
            response = requests.post(
                issue_url,
                headers=self.auth_headers,
                data=json.dumps(params),
                timeout=10,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            try:
                errors_bm = GitHubErrors(**http_error.response.json())
                error_message = errors_bm.message
            except Exception:  # pylint: disable=broad-exception-caught
                error_message = "<Failed to parse GitHub API error response>"
            raise TokenIssueError(error_message) from http_error

        try:
            access_token_bm = AccessToken(**response.json())
        except ValidationError as validation_error:
            error_message = "<Failed to parse Token Issue API response>"
            raise TokenIssueError(error_message) from validation_error

        access_token: TokenResponse = {
            "access_token": access_token_bm.token,
            "expires_at": access_token_bm.expires_at,
            "permissions": access_token_bm.permissions.model_dump(exclude_unset=True),
        }

        if access_token_bm.repositories is not None:
            access_token["repositories"] = sorted(
                [repo.name for repo in access_token_bm.repositories]
            )

        return access_token



"""Package info, etc"""

import importlib.metadata

from .entry import import_app_key, issue_access_token
from .gh import TokenResponse

__all__ = ["import_app_key", "issue_access_token", "TokenResponse"]

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"



"""Top-level functions"""

from .gh import GitHubApp, TokenResponse
from .vault import VaultTransit


def import_app_key(
    # pylint: disable=too-many-arguments
    pem_key: bytes | str,
    key_name: str,
    vault_addr: str,
    vault_token: str,
    transit_backend: str = "transit",
    revoke_vault_token: bool = False,
) -> None:
    """
    Import GitHub App key into Vault's Transit engine

    :param pem_key: The App's PEM formated private RSA key.
    :param key_name: Name which Vault's Transit Engine will know the key by.
    :param vault_addr: Vault instance VAULT_ADDR.
    :param vault_token: Vault instance VAULT_TOKEN.
    :param transit_backend: Transit backend mount path. Defaults to "transit".
    :param revoke_vault_token: Revoke `vault_token` once done? Defaults to False.
    """

    if isinstance(pem_key, str):
        pem_key = pem_key.encode()

    transit = VaultTransit(vault_addr, vault_token, transit_backend)
    transit.import_key(key_name, pem_key)

    if revoke_vault_token:
        transit.revoke_token()


def issue_access_token(
    # pylint: disable=too-many-arguments
    key_name: str,
    vault_addr: str,
    vault_token: str,
    app_id: int | str,
    account: str,
    permissions: None | dict[str, str] = None,
    repositories: None | list[str] = None,
    transit_backend: str = "transit",
    revoke_vault_token: bool = False,
) -> TokenResponse:
    """
    Issue GitHub Access Token

    :param key_name: Name which Vault's Transit Engine knows the App key by.
    :param vault_addr: Vault instance VAULT_ADDR.
    :param vault_token: Vault instance VAULT_TOKEN.
    :param app_id: GitHub App ID.
    :param account: GitHub account to access, where the App is installed.
    :param permissions: Optionally scope (down) token permissions.
    :param repositories: Optionally limit accessible repositories.
    :param transit_backend: Vault Transit backend mount path. Defaults to "transit".
    :param revoke_vault_token: Revoke `vault_token` once done? Defaults to False.

    :return: The requested access token; together with its expiry
        time, permission scope and optionally covered repositories.
    """

    if isinstance(app_id, int):
        app_id = str(app_id)

    transit = VaultTransit(vault_addr, vault_token, transit_backend)
    jwt: str = transit.sign_jwt(key_name, app_id)

    ghapp = GitHubApp(account, jwt)
    access_token: TokenResponse = ghapp.issue_token(permissions, repositories)

    if revoke_vault_token:
        transit.revoke_token()

    return access_token



"""Helper functions"""

import base64
import json
import os
from datetime import datetime

from cryptography.hazmat.primitives import hashes, keywrap, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa


def b64str(data: bytes | str, urlsafe: bool = False) -> str:
    """
    A more type helpful b64encode wrapper function

    :param data: Input to base64 encode
    :param urlsafe: Whatever to make the output JWT URL-safe or not

    :return: Base64 encoded result
    """
    if isinstance(data, str):
        data = data.encode()

    if urlsafe:
        return base64.urlsafe_b64encode(data).decode().rstrip("=")
    return base64.b64encode(data).decode()


def prepare_gh_app_jwt(app_id: str, now: datetime) -> str:
    """
    Prepares the JWT payload needed to authenticate as a GitHub App

    :param app_id: The GitHub App's ID
    :param now: Current UTC time

    :return: The header-and-claims part of a JWT token
    """
    timestamp = int(now.timestamp())
    expire = timestamp + 60

    header = {
        "alg": "RS256",
        "typ": "JWT",
    }

    claims = {
        "iat": timestamp,
        "exp": expire,
        "iss": app_id,
    }

    b64_header = b64str(json.dumps(header), urlsafe=True)
    b64_claims = b64str(json.dumps(claims), urlsafe=True)

    header_and_claims = b64_header + "." + b64_claims
    return header_and_claims


def private_pem_to_der(private_pem: bytes) -> bytes:
    """
    Converts private key from PEM to DER

    :param private_pem: PEM encoded private key

    :return: DER encoded private key
    """
    private_key = serialization.load_pem_private_key(private_pem, password=None)
    private_der = private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return private_der


def vault_wrap_key(der_app_key: bytes, wrapping_key: rsa.RSAPublicKey) -> str:
    """
    Wrap a private key the way Vault wants it wrapped

    :param der_app_key: DER encoded private key
    :param wrapping_key: Vault provided wrapping key

    :return: Wrapped key
    """

    aes_key = os.getrandom(32)

    wrapped_app_key = keywrap.aes_key_wrap_with_padding(aes_key, der_app_key)
    wrapped_aes_key = wrapping_key.encrypt(
        aes_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    wrapped_b64 = b64str(wrapped_aes_key + wrapped_app_key)
    return wrapped_b64
